# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa


# x: Original input string.
# y: Processed string. For simplicity, y is only genreated by deleting some
#    characters from x
# return: marks of original x to indicate whether characters being removed or
# not.
def lcs_marks(x, y):
    nx = len(x)
    ny = len(y)

    # one-based.
    # nx * ny matrix.
    dp = [[None] * (ny + 1) for _ in range(nx + 1)]

    TO_TOP = 1
    TO_TOP_LEFT = 2

    for xi in range(nx + 1):
        for yi in range(min(xi + 1, ny + 1)):
            if xi == 0 or yi == 0:
                dp[xi][yi] = None
                continue
            if x[xi - 1] == y[yi - 1]:
                dp[xi][yi] = TO_TOP_LEFT
            else:
                dp[xi][yi] = TO_TOP

    marks = [False] * nx
    xi = nx
    yi = ny
    while yi > 0 and xi > 0:
        while dp[xi][yi] == TO_TOP:
            xi -= 1
        if dp[xi][yi] is None:
            raise RuntimeError
        # find TO_TOP_LEFT.
        marks[xi - 1] = True
        xi -= 1
        yi -= 1

    return marks



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import re

from mistune import Markdown
from zhlint.mistune_patch import (
    HackedRenderer, HackedBlockLexer, count_newlines,
)
from zhlint.utils import TextElement


def remove_block(pattern, text):
    segments = []
    begin = 0

    for m in re.finditer(pattern, text, flags=re.DOTALL | re.UNICODE):
        segments.append(text[begin:m.start()])
        segments.append('\n' * count_newlines(m))
        begin = m.end()

    if begin < len(text):
        segments.append(text[begin:])

    return ''.join(segments)


def generate_text_elements(text):
    MARK_PATTERN = (
        '<<DOCLINT: TYPE=(.+?), LOC_BEGIN=(\d+), LOC_END=(\d+)>>\n'
    )

    begin = 0
    elements = []

    for m in re.finditer(MARK_PATTERN, text):
        if begin != m.start():
            block_type = m.group(1)
            loc_begin = m.group(2)
            loc_end = m.group(3)

            content = text[begin:m.start()]

            if content.strip():
                elements.append(
                    TextElement(block_type, loc_begin, loc_end, content),
                )

        begin = m.end()

    return elements


def transform(text):

    # yaml header.
    text = remove_block(r'^\s*\-{3}.*?\-{3}', text)

    # latex-block. MUST be executed BEFORE latex-inline.
    text = remove_block(r'\$\$.+?\$\$', text)
    text = remove_block(r'\\\[.+?\\\]', text)

    # latex-inline.
    text = re.sub(r'\$.+?\$', '$$', text)
    text = re.sub(r'\\\(.+?\\\)', '\\\\\(\\\\\)', text)

    # manually add EOF.
    text += '\nEOF\n'

    hacked_md = Markdown(renderer=HackedRenderer(), block=HackedBlockLexer())
    text = hacked_md(text)

    return generate_text_elements(text)



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from zhlint.utils import TextElement
from zhlint.error_detection import (
    process_errors,
)
from zhlint.error_display import ErrorDisplayHandler


def test_process_errors():

    display_handler = ErrorDisplayHandler()
    te = TextElement(
        '', '1', '2',
        '好的句子，好的句子',
    )
    process_errors(display_handler, te)
    assert not display_handler.detected_error

    display_handler = ErrorDisplayHandler()
    te = TextElement(
        '', '1', '2',
        'app坏的句子， 坏的句子',
    )
    process_errors(display_handler, te)
    assert display_handler.detected_error



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import collections as abc
from io import StringIO
from os.path import expanduser


def load_file(fpath):
    fpath = expanduser(fpath)
    with open(fpath, encoding='utf-8') as fin:
        return fin.read()


def try_invoke(inst, method_name):
    method = getattr(inst, method_name, None)
    if callable(method):
        method()


def text2lines(text):
    with StringIO(text) as sin:
        return sin.readlines()


def safelen(line):
    return 0 if not isinstance(line, abc.Sized) else len(line)


def count_newlines(e):
    if isinstance(e, str):
        return count_newlines_in_str(e)
    else:
        return count_newlines_in_match(e)


def count_newlines_in_str(text):
    count = 0
    for c in text:
        if c == '\n':
            count += 1
    return count


def count_newlines_in_match(m):
    return count_newlines_in_str(m.group(0))


def count_offset(prefix):
    offset = 0
    for c in reversed(prefix):
        if c == '\n':
            break
        else:
            offset += 1
    return offset


class TextElement(object):

    def __init__(self, block_type, loc_begin, loc_end, content, offset=None):
        self.block_type = block_type
        self.loc_begin = int(loc_begin)
        self.loc_end = int(loc_end)
        self.content = content
        self.offset = offset



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

import os.path

from zhlint.preprocessor import transform


DATA = os.path.join(
    os.path.dirname(__file__),
    'data',
)


def load_test_md(name):
    return open(os.path.join(DATA, name), encoding='utf-8').read()


def eof(element):
    assert 'EOF\n' == element.content


def test_latex_inline():
    elements = transform(load_test_md('latex_inline.md'))
    for e in elements:
        print(repr(e.content))
    assert 'a line with $$ words.\n' == elements[0].content
    assert 'a line with \\(\\) words.\n' == elements[1].content
    assert '会使 $$ 加入到 $$ 中\n' == elements[2].content
    eof(elements[3])


def test_latex_block():
    elements = transform(load_test_md('latex_block.md'))
    assert 'block 1\n' == elements[0].content
    assert 'block 2\n' == elements[1].content
    eof(elements[2])


def test_yaml_header():
    elements = transform(load_test_md('yaml_header.md'))
    assert 'test line.\n' == elements[0].content
    eof(elements[1])


def test_loc():

    def assert_loc(begin, end, element):
        assert begin == int(element.loc_begin)
        assert end == int(element.loc_end)

    elements = transform(load_test_md('loc.md'))

    assert_loc(5, 5, elements[0])
    assert_loc(9, 11, elements[1])
    assert_loc(16, 16, elements[2])
    eof(elements[3])



# -*- coding: utf-8 -*-
# from __future__ import (
#     division, absolute_import, print_function, unicode_literals,
# )
# from builtins import *                  # noqa
# from future.builtins.disabled import *  # noqa


from mistune import Renderer, BlockLexer
from zhlint.utils import count_newlines


def fmt(text):
    return '%s\n' % text.rstrip('\n')


class HackedRenderer(Renderer):

    def placeholder(self):
        return ''

    def block_code(self, code, lang=None):
        return ''

    def block_quote(self, text):
        return fmt(text)

    def block_html(self, html):
        return ''

    def header(self, text, level, raw=None):
        return text

    def hrule(self):
        return ''

    def list(self, body, ordered=True):
        return body

    def list_item(self, text):
        return fmt(text)

    def paragraph(self, text):
        return fmt(text)

    def table(self, header, body):
        return (
            '%s'
            '%s'
        ) % (header, body)

    def table_row(self, content):
        return content

    def table_cell(self, content, **flags):
        return fmt(content)

    def double_emphasis(self, text):
        return text

    def emphasis(self, text):
        return text

    def codespan(self, text):
        return ''

    def linebreak(self):
        return ''

    def strikethrough(self, text):
        return text

    def text(self, text):
        return text

    def escape(self, text):
        return text

    def autolink(self, link, is_email=False):
        return link

    def link(self, link, title, text):
        return text

    def image(self, src, title, text):
        return ''

    def inline_html(self, html):
        return ''

    def newline(self):
        return ''

    def footnote_ref(self, key, index):
        return ''

    def footnote_item(self, key, text):
        return ''

    def footnotes(self, text):
        return ''


class HackedBlockLexer(BlockLexer):

    def __init__(self, *args, **kwargs):
        super(HackedBlockLexer, self).__init__(*args, **kwargs)
        self.loc = 1
        self.recursive_level = -1

    def parse(self, text, rules=None):
        self.recursive_level += 1

        text = text.rstrip('\n')

        if not rules:
            rules = self.default_rules

        def manipulate(text):
            for key in rules:
                rule = getattr(self.rules, key)
                m = rule.match(text)
                if not m:
                    continue
                getattr(self, 'parse_%s' % key)(m)
                return key, m
            return False, False  # pragma: no cover

        def inject_loc_and_block_type(key, m):
            newlines = count_newlines(m)

            loc_begin = self.loc
            self.loc += newlines
            loc_end = self.loc

            content = m.group(0)
            i = 0
            n = len(content)
            while i < n and content[i] == '\n':
                i += 1
                loc_begin += 1
            i = n - 1
            while i >= 0 and content[i] == '\n':
                i -= 1
                loc_end -= 1

            template = '<<DOCLINT: TYPE={0}, LOC_BEGIN={1}, LOC_END={2}>>\n'
            line = template.format(
                key, str(loc_begin), str(loc_end),
            )
            self.tokens.append({
                'type': 'paragraph',
                'text': line,
            })

        while text:
            key, m = manipulate(text)
            if m is not False:
                text = text[len(m.group(0)):]
                # only inject for top level block.
                if self.recursive_level == 0:
                    inject_loc_and_block_type(key, m)

                continue

            if text:  # pragma: no cover
                raise RuntimeError('Infinite loop at: %s' % text)

        self.recursive_level -= 1
        return self.tokens



# -*- coding: utf-8 -*-
from __future__ import (
    division, absolute_import, print_function, unicode_literals,
)
from builtins import *                  # noqa
from future.builtins.disabled import *  # noqa

from operator import attrgetter

from zhlint.utils import TextElement
from zhlint.error_detection import (
    detect_e101,
    detect_e102,
    detect_e103,
    detect_e104,
    detect_e203,
    detect_e205,
    detect_e206,
    detect_e207,
    detect_e301,

    detect_e201,
    detect_e202,
    detect_e204,

    split_text_element,
)


def check_texts(inputs, detector, should_detected):
    elements = []

    for i in inputs:
        if isinstance(i, str):
            elements.append(
                TextElement('', '1', '2', i)
            )
        elif isinstance(i, TextElement):
            elements.append(i)
        else:
            raise RuntimeError

    for element in elements:
        matches = detector(element)
        if should_detected:
            assert matches
            assert list(matches)
        else:
            assert not matches or not list(matches)


def test_e101():

    check_texts(
        [
            '中文english',
            'english中文',
            '中文  english',
            'english  中文',
            '中文\tenglish',
        ],
        detect_e101,
        should_detected=True,
    )

    check_texts(
        [
            '中文 english',
            'english 中文',
            '中文\nenglish',
            '42\n中文',
        ],
        detect_e101,
        should_detected=False,
    )


def test_e102():

    check_texts(
        [
            '中文1',
            '42中文',
            '中文  42',
            '42  中文',
            '中文\t42',
        ],
        detect_e102,
        should_detected=True,
    )

    check_texts(
        [
            '中文 42',
            '42 中文',
            '中文\n42',
            '中文 \n42',
            '42\n中文',
        ],
        detect_e102,
        should_detected=False,
    )


def test_e103():

    check_texts(
        [
            '42μ',
            '42  μ',
        ],
        detect_e103,
        should_detected=True,
    )

    check_texts(
        [
            '42 μ',
            'μ\n42',
            '42\nμ',
            '42x',
            '42n',
            '42％',
            '42%',
            '42℃',
            '，42',
            '42。',
            'Q3',
            '136-4321-1234',
        ],
        detect_e103,
        should_detected=False,
    )


def test_e104():

    check_texts(
        [
            '(42）',
            '（42)',
            '（42）',
            '42(42)',
            '42  (42)',
            ' (42)',
        ],
        detect_e104,
        should_detected=True,
    )

    check_texts(
        [
            '42 (42)',
            '42\n(42)',
            '(42)',
        ],
        detect_e104,
        should_detected=False,
    )


def test_e203():

    check_texts(
        [
            '中文， 测试',
            '中文 。测试',
            '「 中文」',
        ],
        detect_e203,
        should_detected=True,
    )

    check_texts(
        [
            '中文，测试',
            '中文；测试',
            '「中文」',
            '中文！\n测试',
        ],
        detect_e203,
        should_detected=False,
    )


def test_e205():

    check_texts(
        [
            '中文...',
            '中文.......',
            '中文。。。',
        ],
        detect_e205,
        should_detected=True,
    )

    check_texts(
        [
            '中文......',
            '中文 test@email.com',
            '中文 +135-1234-5678'
            '中文 (1) '
        ],
        detect_e205,
        should_detected=False,
    )


def test_e206():

    check_texts(
        [
            '中文!!',
            '中文！！',
            '中文!！',
            '中文??',
            '中文？？',
            '中文？?',
        ],
        detect_e206,
        should_detected=True,
    )

    check_texts(
        [
            '中文!',
            '中文！',
            '中文?',
            '中文？',
            'english, with space',
        ],
        detect_e206,
        should_detected=False,
    )


def test_e207():

    check_texts(
        [
            '中文~',
        ],
        detect_e207,
        should_detected=True,
    )

    check_texts(
        [
            '中文',
        ],
        detect_e207,
        should_detected=False,
    )


def test_e301():

    check_texts(
        [
            'APP',
            'app',
            'android',
            'ios',
            'IOS',
            'IPHONE',
            'iphone',
            'AppStore',
            'app store',
            'wifi',
            'Wifi',
            'Wi-fi',
            'E-mail',
            'Email',
            'PS',
            'ps',
            'Ps.',

            '我app store我',
        ],
        detect_e301,
        should_detected=True,
    )


def test_e201():

    check_texts(
        [
            '有中文, 错误.',
            'LaTeX 公式 $$.',
            'LaTeX 公式,$$',
            'LaTeX 公式 \(\).',
            'LaTeX 公式,\(\)',
        ],
        detect_e201,
        should_detected=True,
    )

    check_texts(
        [
            '有中文，正确。',
            '有中文，正确......',
            'pure english, nothing wrong.',
            'P.S. 这是一行中文。',
            'LaTeX 公式 $$',
            'LaTeX 公式 \(\)',
        ],
        detect_e201,
        should_detected=False,
    )


def test_e202():

    check_texts(
        [
            'pure english，nothing wrong.',
        ],
        detect_e202,
        should_detected=True,
    )

    check_texts(
        [
            'pure english, nothing wrong.',
        ],
        detect_e202,
        should_detected=False,
    )


def test_e204():

    check_texts(
        [
            # "中文'测试'",
            # '中文"测试"',
            '中文‘测试’',
            '中文“测试”',
        ],
        detect_e204,
        should_detected=True,
    )

    check_texts(
        [
            "中文「测试」",
        ],
        detect_e204,
        should_detected=False,
    )


def test_split_text_element():

    def access(key, te):
        return list(map(attrgetter(key), split_text_element(te)))

    def help_(content, begin, end):
        te = TextElement(
            'paragraph', begin, end,
            content,
        )

        lines = access('content', te)
        begins = access('loc_begin', te)
        ends = access('loc_end', te)
        offsets = access('offset', te)

        return lines, begins, ends, offsets

    content = '''a
b.
c. inline!
d
e.'''

    lines, begins, ends, offsets = help_(content, '1', '5')

    assert ['a\nb.\n', 'c.', ' inline!\n', 'd\ne.'] == lines
    assert [1, 3, 3, 4] == begins
    assert [2, 3, 3, 5] == ends
    assert [0, 0, 2, 0] == offsets

    content = '''a......
b!
c;
d.
e?
f！
g；
h。
i？'''

    lines, begins, ends, offsets = help_(content, '1', '9')
    assert len(lines) == 9

    content = '''P.S. this is a line!'''

    lines, begins, ends, offsets = help_(content, '1', '9')
    assert len(lines) == 1

    content = '''sentence 1.
   （中文！中文！）
sentence 2.'''
    lines, begins, ends, offsets = help_(content, '1', '3')

    assert ['sentence 1.\n', '   （中文！中文！）\nsentence 2.'] == lines
    assert [1, 2] == begins
    assert [1, 3] == ends
    assert [0, 0] == offsets

    content = '''sentence 1.
sentence 2.  sentence 3'''
    lines, begins, ends, offsets = help_(content, '1', '2')

    assert ['sentence 1.\n', 'sentence 2.', '  sentence 3'] == lines
    assert [1, 2, 2] == begins
    assert [1, 2, 2] == ends
    assert [0, 0, 11] == offsets
